@@ -24,8 +24,8 @@
 //! int dim_1d = NagaDimensions1D(image_1d);
 //! ```
 
-use super::{super::FunctionCtx, BackendResult};
-use crate::arena::Handle;
+use super::{super::FunctionCtx, BackendResult, Error};
+use crate::{arena::Handle, proc::NameKey};
 use std::fmt::Write;
 
 #[derive(Clone, Copy, Debug, Hash, Eq, Ord, PartialEq, PartialOrd)]
@@ -41,6 +41,11 @@ pub(super) struct WrappedImageQuery {
     pub(super) query: ImageQuery,
 }
 
+#[derive(Clone, Copy, Debug, Hash, Eq, Ord, PartialEq, PartialOrd)]
+pub(super) struct WrappedConstructor {
+    pub(super) ty: Handle<crate::Type>,
+}
+
 /// HLSL backend requires its own `ImageQuery` enum.
 ///
 /// It is used inside `WrappedImageQuery` and should be unique per ImageQuery function.
@@ -342,6 +347,77 @@ impl<'a, W: Write> super::Writer<'a, W> {
         Ok(())
     }
 
+    pub(super) fn write_wrapped_constructor_function_name(
+        &mut self,
+        constructor: WrappedConstructor,
+    ) -> BackendResult {
+        let name = &self.names[&NameKey::Type(constructor.ty)];
+        write!(self.out, "Construct{}", name)?;
+        Ok(())
+    }
+
+    /// Helper function that write wrapped function for `Expression::Compose` for structures.
+    pub(super) fn write_wrapped_constructor_function(
+        &mut self,
+        module: &crate::Module,
+        constructor: WrappedConstructor,
+    ) -> BackendResult {
+        use crate::back::INDENT;
+
+        const ARGUMENT_VARIABLE_NAME: &str = "arg";
+        const RETURN_VARIABLE_NAME: &str = "ret";
+
+        // Write function return type and name
+        let struct_name = &self.names[&NameKey::Type(constructor.ty)];
+        write!(self.out, "{} ", struct_name)?;
+        self.write_wrapped_constructor_function_name(constructor)?;
+
+        // Write function parameters
+        write!(self.out, "(")?;
+        let members = match module.types[constructor.ty].inner {
+            crate::TypeInner::Struct { ref members, .. } => members,
+            _ => return Err(Error::Unimplemented("non-struct constructor".to_string())),
+        };
+        for (i, member) in members.iter().enumerate() {
+            if i != 0 {
+                write!(self.out, ", ")?;
+            }
+            self.write_type(module, member.ty)?;
+            write!(self.out, " {}{}", ARGUMENT_VARIABLE_NAME, i)?;
+            if let crate::TypeInner::Array { size, .. } = module.types[member.ty].inner {
+                self.write_array_size(module, size)?;
+            }
+        }
+        // Write function body
+        writeln!(self.out, ") {{")?;
+
+        let struct_name = &self.names[&NameKey::Type(constructor.ty)];
+        writeln!(
+            self.out,
+            "{}{} {};",
+            INDENT, struct_name, RETURN_VARIABLE_NAME
+        )?;
+        for i in 0..members.len() as u32 {
+            let field_name = &self.names[&NameKey::StructMember(constructor.ty, i)];
+            //TODO: handle arrays?
+            writeln!(
+                self.out,
+                "{}{}.{} = {}{};",
+                INDENT, RETURN_VARIABLE_NAME, field_name, ARGUMENT_VARIABLE_NAME, i,
+            )?;
+        }
+
+        // Write return value
+        writeln!(self.out, "{}return {};", INDENT, RETURN_VARIABLE_NAME)?;
+
+        // End of function body
+        writeln!(self.out, "}}")?;
+        // Write extra new line
+        writeln!(self.out)?;
+
+        Ok(())
+    }
+
     /// Helper function that write wrapped function for `Expression::ImageQuery` and `Expression::ArrayLength`
     ///
     /// <https://docs.microsoft.com/en-us/windows/win32/direct3dhlsl/dx-graphics-hlsl-to-getdimensions>
@@ -371,9 +447,9 @@ impl<'a, W: Write> super::Writer<'a, W> {
                         writable: storage_access.contains(crate::StorageAccess::STORE),
                     };
 
-                    if !self.wrapped_array_lengths.contains(&wal) {
+                    if !self.wrapped.array_lengths.contains(&wal) {
                         self.write_wrapped_array_length_function(module, wal, handle, func_ctx)?;
-                        self.wrapped_array_lengths.insert(wal);
+                        self.wrapped.array_lengths.insert(wal);
                     }
                 }
                 crate::Expression::ImageQuery { image, query } => {
@@ -391,9 +467,19 @@ impl<'a, W: Write> super::Writer<'a, W> {
                         _ => unreachable!("we only query images"),
                     };
 
-                    if !self.wrapped_image_queries.contains(&wiq) {
+                    if !self.wrapped.image_queries.contains(&wiq) {
                         self.write_wrapped_image_query_function(module, wiq, handle, func_ctx)?;
-                        self.wrapped_image_queries.insert(wiq);
+                        self.wrapped.image_queries.insert(wiq);
+                    }
+                }
+                crate::Expression::Compose { ty, components: _ } => {
+                    let constructor = match module.types[ty].inner {
+                        crate::TypeInner::Struct { .. } => WrappedConstructor { ty },
+                        _ => continue,
+                    };
+                    if !self.wrapped.constructors.contains(&constructor) {
+                        self.write_wrapped_constructor_function(module, constructor)?;
+                        self.wrapped.constructors.insert(constructor);
                     }
                 }
                 _ => {}
