@@ -219,67 +219,27 @@ impl super::Instruction {
         instruction
     }
 
+    #[allow(clippy::too_many_arguments)]
     pub(super) fn type_image(
         id: Word,
         sampled_type_id: Word,
         dim: spirv::Dim,
-        arrayed: bool,
-        image_class: crate::ImageClass,
+        flags: super::ImageTypeFlags,
+        image_format: spirv::ImageFormat,
     ) -> Self {
         let mut instruction = Self::new(Op::TypeImage);
         instruction.set_result(id);
         instruction.add_operand(sampled_type_id);
         instruction.add_operand(dim as u32);
-
-        let (depth, multi, sampled) = match image_class {
-            crate::ImageClass::Sampled { kind: _, multi } => (false, multi, true),
-            crate::ImageClass::Depth { multi } => (true, multi, true),
-            crate::ImageClass::Storage { .. } => (false, false, false),
-        };
-        instruction.add_operand(depth as u32);
-        instruction.add_operand(arrayed as u32);
-        instruction.add_operand(multi as u32);
-        instruction.add_operand(if sampled { 1 } else { 2 });
-
-        let format = match image_class {
-            crate::ImageClass::Storage { format, .. } => match format {
-                crate::StorageFormat::R8Unorm => spirv::ImageFormat::R8,
-                crate::StorageFormat::R8Snorm => spirv::ImageFormat::R8Snorm,
-                crate::StorageFormat::R8Uint => spirv::ImageFormat::R8ui,
-                crate::StorageFormat::R8Sint => spirv::ImageFormat::R8i,
-                crate::StorageFormat::R16Uint => spirv::ImageFormat::R16ui,
-                crate::StorageFormat::R16Sint => spirv::ImageFormat::R16i,
-                crate::StorageFormat::R16Float => spirv::ImageFormat::R16f,
-                crate::StorageFormat::Rg8Unorm => spirv::ImageFormat::Rg8,
-                crate::StorageFormat::Rg8Snorm => spirv::ImageFormat::Rg8Snorm,
-                crate::StorageFormat::Rg8Uint => spirv::ImageFormat::Rg8ui,
-                crate::StorageFormat::Rg8Sint => spirv::ImageFormat::Rg8i,
-                crate::StorageFormat::R32Uint => spirv::ImageFormat::R32ui,
-                crate::StorageFormat::R32Sint => spirv::ImageFormat::R32i,
-                crate::StorageFormat::R32Float => spirv::ImageFormat::R32f,
-                crate::StorageFormat::Rg16Uint => spirv::ImageFormat::Rg16ui,
-                crate::StorageFormat::Rg16Sint => spirv::ImageFormat::Rg16i,
-                crate::StorageFormat::Rg16Float => spirv::ImageFormat::Rg16f,
-                crate::StorageFormat::Rgba8Unorm => spirv::ImageFormat::Rgba8,
-                crate::StorageFormat::Rgba8Snorm => spirv::ImageFormat::Rgba8Snorm,
-                crate::StorageFormat::Rgba8Uint => spirv::ImageFormat::Rgba8ui,
-                crate::StorageFormat::Rgba8Sint => spirv::ImageFormat::Rgba8i,
-                crate::StorageFormat::Rgb10a2Unorm => spirv::ImageFormat::Rgb10a2ui,
-                crate::StorageFormat::Rg11b10Float => spirv::ImageFormat::R11fG11fB10f,
-                crate::StorageFormat::Rg32Uint => spirv::ImageFormat::Rg32ui,
-                crate::StorageFormat::Rg32Sint => spirv::ImageFormat::Rg32i,
-                crate::StorageFormat::Rg32Float => spirv::ImageFormat::Rg32f,
-                crate::StorageFormat::Rgba16Uint => spirv::ImageFormat::Rgba16ui,
-                crate::StorageFormat::Rgba16Sint => spirv::ImageFormat::Rgba16i,
-                crate::StorageFormat::Rgba16Float => spirv::ImageFormat::Rgba16f,
-                crate::StorageFormat::Rgba32Uint => spirv::ImageFormat::Rgba32ui,
-                crate::StorageFormat::Rgba32Sint => spirv::ImageFormat::Rgba32i,
-                crate::StorageFormat::Rgba32Float => spirv::ImageFormat::Rgba32f,
-            },
-            _ => spirv::ImageFormat::Unknown,
-        };
-
-        instruction.add_operand(format as u32);
+        instruction.add_operand(flags.contains(super::ImageTypeFlags::DEPTH) as u32);
+        instruction.add_operand(flags.contains(super::ImageTypeFlags::ARRAYED) as u32);
+        instruction.add_operand(flags.contains(super::ImageTypeFlags::MULTISAMPLED) as u32);
+        instruction.add_operand(if flags.contains(super::ImageTypeFlags::SAMPLED) {
+            1
+        } else {
+            2
+        });
+        instruction.add_operand(image_format as u32);
         instruction
     }
 
@@ -932,3 +892,55 @@ impl super::Instruction {
         instruction
     }
 }
+
+impl From<crate::StorageFormat> for spirv::ImageFormat {
+    fn from(format: crate::StorageFormat) -> Self {
+        use crate::StorageFormat as Sf;
+        match format {
+            Sf::R8Unorm => Self::R8,
+            Sf::R8Snorm => Self::R8Snorm,
+            Sf::R8Uint => Self::R8ui,
+            Sf::R8Sint => Self::R8i,
+            Sf::R16Uint => Self::R16ui,
+            Sf::R16Sint => Self::R16i,
+            Sf::R16Float => Self::R16f,
+            Sf::Rg8Unorm => Self::Rg8,
+            Sf::Rg8Snorm => Self::Rg8Snorm,
+            Sf::Rg8Uint => Self::Rg8ui,
+            Sf::Rg8Sint => Self::Rg8i,
+            Sf::R32Uint => Self::R32ui,
+            Sf::R32Sint => Self::R32i,
+            Sf::R32Float => Self::R32f,
+            Sf::Rg16Uint => Self::Rg16ui,
+            Sf::Rg16Sint => Self::Rg16i,
+            Sf::Rg16Float => Self::Rg16f,
+            Sf::Rgba8Unorm => Self::Rgba8,
+            Sf::Rgba8Snorm => Self::Rgba8Snorm,
+            Sf::Rgba8Uint => Self::Rgba8ui,
+            Sf::Rgba8Sint => Self::Rgba8i,
+            Sf::Rgb10a2Unorm => Self::Rgb10a2ui,
+            Sf::Rg11b10Float => Self::R11fG11fB10f,
+            Sf::Rg32Uint => Self::Rg32ui,
+            Sf::Rg32Sint => Self::Rg32i,
+            Sf::Rg32Float => Self::Rg32f,
+            Sf::Rgba16Uint => Self::Rgba16ui,
+            Sf::Rgba16Sint => Self::Rgba16i,
+            Sf::Rgba16Float => Self::Rgba16f,
+            Sf::Rgba32Uint => Self::Rgba32ui,
+            Sf::Rgba32Sint => Self::Rgba32i,
+            Sf::Rgba32Float => Self::Rgba32f,
+        }
+    }
+}
+
+impl From<crate::ImageDimension> for spirv::Dim {
+    fn from(dim: crate::ImageDimension) -> Self {
+        use crate::ImageDimension as Id;
+        match dim {
+            Id::D1 => Self::Dim1D,
+            Id::D2 => Self::Dim2D,
+            Id::D3 => Self::Dim3D,
+            Id::Cube => Self::DimCube,
+        }
+    }
+}
